import sys
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
import dask.array as da

from sigma import grib2sigma


######################
### NEEDED MODULES ###
######################

# there are only available in MN5 (load them in the same order)
"""
module purge
module load netcdf4-python/1.6.5-foss-2023b
module load xarray/2024.5.0-gfbf-2023b-Python-3.11.5
module load dask/2024.5.1-gfbf-2023b-Python-3.11.5
module load cfgrib/0.9.15.0-foss-2023b-Python-3.11.5
module load ecCodes/2.35.1-gompi-2023b
"""


def grib2amoc(grib_file, var_name, stf_name, area_file, sigma_values=None,
              sigma_levels=None, sigma_name=None, save_path=None):
    """
    Convert grib file to netCDF

    Parameters
    ----------
    grib_file: str or pathlib.Path
        Path to the file to be converted with the data in the original
        grid from Nemo.
    save_path: str or pathlib.Path
        Path to save the netcdf file.
    var_name: str
        Name of the variable in the grid_file, should be defined in
        var_defs.
    mesh_file: str or pathlib.Path
        Path to the Nemo mesh file wich contins level, latitude, and
        longitude data.

    Returns
    -------
    None

    """
    ###################
    ### GET VAR DEF ###
    ###################

    # grid shape
    y_len, x_len = 3606, 4322

    # units changes
    m3s_to_sverdrup = 1e-6

    # data in mesh file
    area_name = "areav"

    #################
    ### OPEN DATA ###
    #################

    ds = xr.open_mfdataset(grib_file, engine="cfgrib")

    ################
    ### FIX TIME ###
    ################

    # rename time coordinate
    ds = ds.rename({"time": "start_time"})

    # get the (time, step) selection fro the average, 1st day of the next month
    valid_time = pd.DatetimeIndex(ds['valid_time'].values.ravel())
    valid_time_sel = (valid_time.day == 1).reshape(ds['valid_time'].shape)

    # get index to extract (time, step) combinations
    step_sel = np.where(valid_time_sel)[-1]
    time_sel = range(len(ds.start_time))

    # define DataArrays for advanced indexing
    # https://docs.xarray.dev/en/stable/user-guide/indexing.html#more-advanced-indexing
    start_time_idx = xr.DataArray(time_sel, dims='t')
    step_idx = xr.DataArray(step_sel, dims='t')

    # select (time, step) combinations
    ds = ds.isel(start_time=start_time_idx, step=step_idx)

    # get start_time, end_time, and avg_time
    start_time = ds['start_time']
    end_time = ds['valid_time']
    time = (start_time+(end_time-start_time)*0.5)

    time_bnds = xr.DataArray(
        np.transpose([start_time, end_time]),
        dims=["time", "bnds"]
    )

    # time and time_bnds are used later to build the final dataset

    #################
    ### OPEN AREA ###
    #################

    with xr.open_mfdataset(area_file) as file:
        area = file[area_name]

    ################
    ### FIX GRID ###
    ################

    # get data
    darray = ds[var_name]

    # reshape data
    data = darray.data.reshape(*darray.shape[:-1], y_len, x_len)
    dims = list(darray.dims[:-1]) + ["j", "i"]

    new_darray = xr.DataArray(data, dims=dims)
    new_darray = new_darray.assign_coords(
        j=("j", np.arange(1, y_len+1)),
        i=("i", np.arange(1, x_len+1)),
    )

    # extract same positions as in area
    new_darray = new_darray.sel(j=area.j, i=area.i)

    #################
    ### TRANSPORT ###
    #################

    # compute transport
    transport = - new_darray.data * area.data[None] * m3s_to_sverdrup

    if sigma_values is None:
        # convert regular array to masked array
        transport = da.ma.masked_where(np.isnan(transport), transport)
        # compute overturning stream function -> sum over x-axis (axis=-1),
        # invert cumsum over z-axis (axis=1)
        stfunc = transport.sum(axis=-1)[:, ::-1].cumsum(axis=1)[:, ::-1]
        # convert masked array to numpy.nan
        stfunc = da.ma.filled(stfunc, np.nan)

        # create xr.DataArray for stream function
        da_stf = xr.DataArray(stfunc, dims=["time", "lev", "j"])
        # assing coords
        da_stf = da_stf.assign_coords(
            time=("time", time.data),
            lev=("lev", area.lev.data),
            j=("j", new_darray.j.data),
            lat=("j", area.lat.data.max(axis=1)),
        )
        da_stf.lev.attrs = dict(
            standard_name="depth",
            long_name="ocean depth coordinate",
            axis="Z",
            positive="down",
            units="m"
        )
    else:
        sigma_m = .5*(sigma_levels[1:]+sigma_levels[:-1])
        sigma_bnds = np.empty((len(sigma_levels), 2), dtype=float)
        sigma_bnds[0, 0] = 1.
        sigma_bnds[-1, 1] = 50.
        sigma_bnds[1:, 0] = sigma_m
        sigma_bnds[:-1, 1] = sigma_m

        transport = transport.compute()
        n_bnds = len(sigma_bnds)
        # get final shape
        shape = list(transport.shape)[:-1]
        shape[1] = n_bnds
        # generate cube to save data
        stfunc = np.empty(shape)
        slice_dd = [slice(None) for i in range(1)] + [None]
        # Set masked values to 0
        transport[np.isnan(transport)] = 0.
        for dd in range(n_bnds):
            # Set upper values to 0 to compute cumulative
            # sum in inverse order
            transport[sigma_values < sigma_bnds[dd, 0]] = 0.
            slice_dd[-1] = dd
            stfunc[*slice_dd] = transport.sum(axis=(1, 3))

        # create xr.DataArray for stream function
        da_stf = xr.DataArray(stfunc, dims=["time", "density_level", "j"])
        # assing coords
        da_stf = da_stf.assign_coords(
            time=("time", time.data),
            density_level=("density_level", sigma_levels),
            j=("j", new_darray.j.data),
            lat=("j", area.lat.data.max(axis=1)),
        )
        da_stf.density_level.attrs = dict(
            standard_name="ocean_sigma_coordinate",
            long_name="density_level_defined_with_%s" % sigma_name,
            axis="Z",
            positive="down",
            units="kg m-3"
        )

    # assign attributes
    da_stf.time.attrs = dict(
        standard_name="time",
        long_name="time",
        axis="T",
        bounds="time_bnds"
    )
    da_stf.lat.attrs = dict(
        standard_name="latitude",
        units="degrees_north"
    )
    da_stf.j.attrs = dict(
        long_name="cell index along second dimension",
        units="1"
    )

    da_stf.attrs = dict(
        standard_name="",
        long_name="Ocean Y Overturning Volume Streamfunction",
        comment="Overturning volume streamfunction arising from "
                "resolved processes.",
        units="Sverdrup"
    )

    # create dataset
    ds_stf = xr.Dataset({stf_name: da_stf, "time_bnds": time_bnds})

    #################
    ### SAVE FILE ###
    #################

    # define encoding
    def_enconding = {"_FillValue": None, "dtype": "float32"}
    encoding = {ivar: def_enconding for ivar in ds_stf.data_vars}
    encoding.update({ivar: def_enconding for ivar in ds_stf.coords})

    # fill value for the var
    encoding[stf_name]["missing_value"] = 1e20
    encoding[stf_name]["_FillValue"] = 1e20

    # define time encoding as attribute
    del encoding["time"]
    del encoding["time_bnds"]

    ds_stf.time.encoding["units"] = "days since 1950-01-01"
    ds_stf.time.encoding["dtype"] = "double"
    ds_stf.time.encoding["_FillValue"] = None
    ds_stf.time_bnds.encoding["units"] = "days since 1950-01-01"
    ds_stf.time_bnds.encoding["dtype"] = "double"
    ds_stf.time_bnds.encoding["_FillValue"] = None

    # save data
    ds_stf.to_netcdf(save_path, unlimited_dims='time', encoding=encoding)


if __name__ == "__main__":

    path = Path(sys.argv[1])
    path_t = (path.parent.parent / 'thetao' / path.stem.replace("vo_", "thetao_")).with_suffix(".grib")
    path_s = (path.parent.parent / 'so' / path.stem.replace("vo_", "so_")).with_suffix(".grib")
    save_path = (path.parent.parent / 'vsftysigma2' /
                 path.stem.replace("vo_", "vsftysigma2_")).with_suffix(".nc")
    save_path.parent.mkdir(exist_ok=True, parents=True)


    # Skip existing files
    if not save_path.exists():
        sigma_values = grib2sigma(
            tgrib_file=path_t,
            sgrib_file=path_s,
            tvar_name="avg_thetao",
            svar_name="avg_so",
            sigma=2,
            area_file="/gpfs/scratch/bsc32/bsc032050/ifsnemo_grib_post/"
                      "production/areav_cutted.nc"
        ).compute()
        grib2amoc(
            grib_file=path,
            save_path=save_path,
            var_name="avg_vo",
            stf_name="vsftysigma2",
            sigma_name="sigma2",
            sigma_levels=np.array([30.3, 30.7, 31.1, 31.5, 31.9, 32.3, 32.7, 33.1, 33.5, 33.9, 34.3, 34.7, 35.05, 35.25, 35.45, 35.65, 35.85, 36.025, 36.125, 36.225, 36.325, 36.425, 36.525, 36.625, 36.725, 36.825, 36.925, 37.025, 37.125, 37.225, 37.325, 37.425, 37.525, 37.625]),
            sigma_values=sigma_values,
            area_file="/gpfs/scratch/bsc32/bsc032050/ifsnemo_grib_post/"
                      "production/areav_cutted.nc"
        )
        print("Saved %s" % save_path)
