import sys
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
import dask.array as da

######################
### NEEDED MODULES ###
######################

# there are only available in MN5 (load them in the same order)
"""
module purge
module load netcdf4-python/1.6.5-foss-2023b
module load xarray/2024.5.0-gfbf-2023b-Python-3.11.5
module load dask/2024.5.1-gfbf-2023b-Python-3.11.5
module load cfgrib/0.9.15.0-foss-2023b-Python-3.11.5
module load ecCodes/2.35.1-gompi-2023b
"""

# Reference salinity
SSO = 35.16504

# Reference potential density (kg m-3)
REF_POTENTIAL_DENSITY = 1000

# Reference pressure per sigma (dbar)
REF_PRESSURE_PER_SIGMA = 1000


def grib2sigma(tgrib_file, sgrib_file, tvar_name, svar_name, sigma,
               save_path=None, mesh_file=None, area_file=None):
    """
    Convert grib file to netCDF

    Parameters
    ----------
    grib_file: str or pathlib.Path
        Path to the file to be converted with the data in the original
        grid from Nemo.
    save_path: str or pathlib.Path
        Path to save the netcdf file.
    var_name: str
        Name of the variable in the grid_file, should be defined in
        var_defs.
    mesh_file: str or pathlib.Path
        Path to the Nemo mesh file wich contins level, latitude, and
        longitude data.

    Returns
    -------
    None

    """
    thetao = _open_grib(tgrib_file, tvar_name)
    so = _open_grib(sgrib_file, svar_name)

    if area_file is not None:
        with xr.open_mfdataset(area_file) as area:
            thetao_dt = thetao['data'].sel(j=area.j, i=area.i)
            so_dt = so['data'].sel(j=area.j, i=area.i)
    else:
        thetao_dt = thetao['data']
        so_dt = so['data']

    if thetao['units'].lower() in ('k', 'kelvin'):
        thetao_dt -= 273.15

    absso = convert_so(so_dt)
    bigthetao = convert_thetao(absso, thetao_dt)

    sigma_name = "sigma%d" % sigma
    # Get reference pressure (dbar) for the given sigma
    ref_pressure = sigma * REF_PRESSURE_PER_SIGMA
    # Compute the values
    sigma_values = compute_density(
        absso,
        bigthetao,
        ref_pressure
    )

    sigma_values -= REF_POTENTIAL_DENSITY

    if save_path is None:
        return sigma_values

    with xr.open_mfdataset(mesh_file).isel(t=0) as file:
        lev = file['nav_lev'].astype('float32', copy=False).data
        lon = file['glamt'].astype('float32', copy=False).data
        lat = file['gphit'].astype('float32', copy=False).data

    da_sigma = xr.DataArray(sigma_values, dims=['time', 'lev', 'j', 'i'])
    da_sigma = da_sigma.assign_coords(
        time=('time', thetao['time']),
        lev=("lev", lev),
        j=('j', np.arange(1, len(da_sigma.j)+1)),
        i=('i', np.arange(1, len(da_sigma.i)+1)),
        lat=(['j', 'i'], lat),
        lon=(['j', 'i'], lon),
    )

    # assign attributes
    da_sigma.time.attrs = dict(
        standard_name='time',
        long_name='time',
        axis='T',
        bounds='time_bnds'
    )
    da_sigma.lev.attrs = dict(
        standard_name="depth",
        long_name="ocean depth coordinate",
        axis="Z",
        positive="down",
        units="m"
    )
    da_sigma.lat.attrs = dict(
        standard_name='latitude',
        units='degrees_north'
    )
    da_sigma.j.attrs = dict(
        long_name='cell index along second dimension',
        units='1'
    )
    da_sigma.attrs = dict(
        standard_name='sea_water_sigma_theta',
        long_name='potential density anomaly (potential density minus '
                  '%d kg/m3) with reference pressure of %d dbar' % (
                      REF_POTENTIAL_DENSITY, ref_pressure
                  ),
        units='kg m-3'
    )

    # create dataset
    ds_sigma = xr.Dataset(
        {sigma_name: da_sigma, 'time_bnds':  thetao['time_bnds']}
    )


    #################
    ### SAVE FILE ###
    #################

    # define encoding
    def_enconding = {'_FillValue': None, 'dtype': 'float32'}
    encoding = {ivar: def_enconding for ivar in ds_sigma.data_vars}
    encoding.update({ivar: def_enconding for ivar in ds_sigma.coords})

    # fill value for the var
    encoding[sigma_name]['missing_value'] = 1e20
    encoding[sigma_name]['_FillValue'] = 1e20

    # define time encoding as attribute
    del encoding['time']
    del encoding['time_bnds']

    ds_sigma.time.encoding['units'] = 'days since 1950-01-01'
    ds_sigma.time.encoding['dtype'] = 'double'
    ds_sigma.time.encoding['_FillValue'] = None
    ds_sigma.time_bnds.encoding['units'] = 'days since 1950-01-01'
    ds_sigma.time_bnds.encoding['dtype'] = 'double'
    ds_sigma.time_bnds.encoding['_FillValue'] = None

    # save data
    ds_sigma.to_netcdf(save_path, unlimited_dims='time', encoding=encoding)


def convert_so(so):
    """
    Convert practical salinity to absolute.

    Parameters
    ----------
    so: iris.cube.Cube
        Cube containing the practical salinity values (psu or 0.001).

    Returns
    -------
    absso: dask.array.core.Array
        Masked array containing the absolute salinity values (g/kg).

    """
    return SSO/35 * so


def convert_thetao(absso, thetao):
    """
    Parameters
    ----------
    absso: dask.array.core.Array
        Masked array containing the absolute salinity values.
    thetao: dask.array.core.Array
        Masked array containing the potential temperature values (degC).

    Returns
    -------
    bigthetao: dask.array.core.Array
        Masked array containing the conservative temperature values (degC).

    Note
    ----
    http://www.teos-10.org/pubs/gsw/html/gsw_CT_from_pt.html

    """
    x = np.sqrt(0.0248826675584615*absso)
    y = thetao*0.025e0
    enthalpy = 61.01362420681071e0 + y*(168776.46138048015e0 +
        y*(-2735.2785605119625e0 + y*(2574.2164453821433e0 +
        y*(-1536.6644434977543e0 + y*(545.7340497931629e0 +
        (-50.91091728474331e0 - 18.30489878927802e0*y)*
        y))))) + x**2*(268.5520265845071e0 + y*(-12019.028203559312e0 +
        y*(3734.858026725145e0 + y*(-2046.7671145057618e0 +
        y*(465.28655623826234e0 + (-0.6370820302376359e0 -
        10.650848542359153e0*y)*y)))) +
        x*(937.2099110620707e0 + y*(588.1802812170108e0 +
        y*(248.39476522971285e0 + (-3.871557904936333e0 -
        2.6268019854268356e0*y)*y)) +
        x*(-1687.914374187449e0 + x*(246.9598888781377e0 +
        x*(123.59576582457964e0 - 48.5891069025409e0*x)) +
        y*(936.3206544460336e0 +
        y*(-942.7827304544439e0 + y*(369.4389437509002e0 +
        (-33.83664947895248e0 - 9.987880382780322e0*y)*y))))))

    # bigthetao = enthalpy/Specific heat
    return enthalpy/3991.86795711963


def compute_density(absso, bigthetao, ref_pressure):
    """
    Computes the potential density in-situ.

    Parameters
    ----------
    absso: dask.array.core.Array
        Masked array containing the absolute salinity values (g/kg).
    bigthetao: dask.array.core.Array
        Masked array containing the conservative temperature values (degC).
    ref_pressure: float
        Reference pressure (dbar).

    Returns
    -------
    rho: dask.array.core.Array
        Masked array containing the potential density in-situ values (kg m-3).

    Note
    ----
    https://github.com/fabien-roquet/polyTEOS/blob/36b9aef6cd2755823b5d3a7349cfe64a6823a73e/polyTEOS10.py#L57

    """
    # reduced variables
    SAu = 40.*SSO/35.
    CTu = 40.
    Zu = 1e4
    deltaS = 32.
    ss = np.sqrt((absso+deltaS)/SAu)
    tt = bigthetao / CTu
    pp = ref_pressure / Zu

    # vertical reference profile of density
    R00 = 4.6494977072e+01
    R01 = -5.2099962525e+00
    R02 = 2.2601900708e-01
    R03 = 6.4326772569e-02
    R04 = 1.5616995503e-02
    R05 = -1.7243708991e-03
    r0 = (((((R05*pp + R04)*pp + R03)*pp + R02)*pp + R01)*pp + R00)*pp

    # density anomaly
    R000 = 8.0189615746e+02
    R100 = 8.6672408165e+02
    R200 = -1.7864682637e+03
    R300 = 2.0375295546e+03
    R400 = -1.2849161071e+03
    R500 = 4.3227585684e+02
    R600 = -6.0579916612e+01
    R010 = 2.6010145068e+01
    R110 = -6.5281885265e+01
    R210 = 8.1770425108e+01
    R310 = -5.6888046321e+01
    R410 = 1.7681814114e+01
    R510 = -1.9193502195e+00
    R020 = -3.7074170417e+01
    R120 = 6.1548258127e+01
    R220 = -6.0362551501e+01
    R320 = 2.9130021253e+01
    R420 = -5.4723692739e+00
    R030 = 2.1661789529e+01
    R130 = -3.3449108469e+01
    R230 = 1.9717078466e+01
    R330 = -3.1742946532e+00
    R040 = -8.3627885467e+00
    R140 = 1.1311538584e+01
    R240 = -5.3563304045e+00
    R050 = 5.4048723791e-01
    R150 = 4.8169980163e-01
    R060 = -1.9083568888e-01
    R001 = 1.9681925209e+01
    R101 = -4.2549998214e+01
    R201 = 5.0774768218e+01
    R301 = -3.0938076334e+01
    R401 = 6.6051753097e+00
    R011 = -1.3336301113e+01
    R111 = -4.4870114575e+00
    R211 = 5.0042598061e+00
    R311 = -6.5399043664e-01
    R021 = 6.7080479603e+00
    R121 = 3.5063081279e+00
    R221 = -1.8795372996e+00
    R031 = -2.4649669534e+00
    R131 = -5.5077101279e-01
    R041 = 5.5927935970e-01
    R002 = 2.0660924175e+00
    R102 = -4.9527603989e+00
    R202 = 2.5019633244e+00
    R012 = 2.0564311499e+00
    R112 = -2.1311365518e-01
    R022 = -1.2419983026e+00
    R003 = -2.3342758797e-02
    R103 = -1.8507636718e-02
    R013 = 3.7969820455e-01

    rz3 = R013*tt + R103*ss + R003
    rz2 = (R022*tt+R112*ss+R012)*tt+(R202*ss+R102)*ss+R002
    rz1 = (((R041*tt+R131*ss+R031)*tt +
            (R221*ss+R121)*ss+R021)*tt +
           ((R311*ss+R211)*ss+R111)*ss+R011)*tt + \
        (((R401*ss+R301)*ss+R201)*ss+R101)*ss+R001
    rz0 = (((((R060*tt+R150*ss+R050)*tt +
              (R240*ss+R140)*ss+R040)*tt +
             ((R330*ss+R230)*ss+R130)*ss+R030)*tt +
            (((R420*ss+R320)*ss+R220)*ss+R120)*ss+R020)*tt +
           ((((R510*ss+R410)*ss+R310)*ss+R210)*ss+R110)*ss+R010)*tt + \
        (((((R600*ss+R500)*ss+R400)*ss+R300)*ss+R200)*ss+R100)*ss+R000
    r = ((rz3*pp + rz2)*pp + rz1)*pp + rz0

    # in-situ density
    rho = r + r0

    return rho


def _open_grib(grib_file, var_name):
    """
    Open grib file to netCDF

    Parameters
    ----------
    grib_file: str or pathlib.Path
        Path to the file to be converted with the data in the original
        grid from Nemo.
    save_path: str or pathlib.Path
        Path to save the netcdf file.
    var_name: str
        Name of the variable in the grid_file, should be defined in
        var_defs.
    mesh_file: str or pathlib.Path
        Path to the Nemo mesh file wich contins level, latitude, and
        longitude data.

    Returns
    -------
    None

    """
    ###################
    ### GET VAR DEF ###
    ###################

    # grid shape
    y_len, x_len = 3606, 4322

    #################
    ### OPEN DATA ###
    #################

    ds = xr.open_mfdataset(grib_file, engine="cfgrib")

    ################
    ### FIX TIME ###
    ################

    # rename time coordinate
    ds = ds.rename({"time": "start_time"})

    # get the (time, step) selection fro the average, 1st day of the next month
    valid_time = pd.DatetimeIndex(ds['valid_time'].values.ravel())
    valid_time_sel = (valid_time.day == 1).reshape(ds['valid_time'].shape)

    # get index to extract (time, step) combinations
    step_sel = np.where(valid_time_sel)[-1]
    time_sel = range(len(ds.start_time))

    # define DataArrays for advanced indexing
    # https://docs.xarray.dev/en/stable/user-guide/indexing.html#more-advanced-indexing
    start_time_idx = xr.DataArray(time_sel, dims='t')
    step_idx = xr.DataArray(step_sel, dims='t')

    # select (time, step) combinations
    ds = ds.isel(start_time=start_time_idx, step=step_idx)

    # get start_time, end_time, and avg_time
    start_time = ds['start_time']
    end_time = ds['valid_time']
    time = start_time+(end_time-start_time)*0.5

    time_bnds = xr.DataArray(
        np.transpose([start_time, end_time]),
        dims=["time", "bnds"]
    )

    # time and time_bnds are used later to build the final dataset

    ################
    ### FIX GRID ###
    ################

    # get data
    darray = ds[var_name]

    # reshape data
    data = darray.data.reshape(*darray.shape[:-1], y_len, x_len)
    dims = list(darray.dims[:-1]) + ["j", "i"]

    new_darray = xr.DataArray(data, dims=dims)
    new_darray = new_darray.assign_coords(
        j=("j", np.arange(1, y_len+1)),
        i=("i", np.arange(1, x_len+1)),
    )

    return dict(data=new_darray, time=time.data, time_bnds=time_bnds,
                units=darray.attrs['units'])


if __name__ == "__main__":

    path_t = Path(
        "/gpfs/scratch/bsc32/bsc032050/ifsnemo_grib_post/data/thetao/"
        "thetao_IFS-CY48R1-Tco1279-NEMO4-eO12-EERIE_hist-1950_r1i1p1f1_gn_195001-195006.grib"
    )
    path_s = (path_t.parent.parent / 'so' / path_t.stem.replace("thetao_", "so_")).with_suffix(".grib")
    save_path = (path_t.parent.parent / 'sigma2' / path_t.stem.replace("thetao_", "sigma2_")).with_suffix(".nc")
    save_path.parent.mkdir(exist_ok=True, parents=True)

    grib2sigma(
        tgrib_file=path_t,
        sgrib_file=path_s,
        tvar_name="avg_thetao",
        svar_name="avg_so",
        sigma=2,
        save_path=save_path,
        mesh_file="/gpfs/scratch/bsc32/bsc032050/ifsnemo_grib_post/production/mesh_NEMO_EERIE.nc"
    )
